﻿module pes.system.locker;

// 
// 無駄なくらいロックするマルチスレッド対応ワンコインロッカー(MAP)
// 
// 
// try-finallyは、
// 代入演算子や比較演算子がオーバーロードされていて
// 例外も考えられるかナーといちおういれておいた
// 
// lock/unlockは
// 長いループ等で明示的にロックしたいときに使う
// ロッカーを開けっ放しでゴソゴソする感じ
// 
// unlockはスレッドIDベースのチェックしかしていないので
// IDが短い期間で再利用されるような環境ではデバッグに注意
// 
// ソートはvaluesで取り出したやつをソートすべし

private import	std.thread;
private import	pes.system.semaphore;
private import 	std.stdio;
private import	pes.hashtable;

debug(locker)
{
	import std.stdio;
}

enum LockerMode
{
	NON,
	AUTO,
}

class Locker(KT, VT)
{
public:
	enum LockType 
	{
		AUTOMATIC,
		MANUAL,
	}
	
protected:
	HashTable!(KT, VT)	m_aa;
	semaphore			m_id;
	SemaphoreResource 	m_resource;
	uint				m_last_locker;
	uint				m_counter;
	
public:
	this(SemaphoreResource resource, semaphore id)
	{
		m_id 		= id;
		m_resource 	= resource;
		m_aa		= new HashTable!(KT, VT);
	}
	
	// manual lock
	void lock()
	out {
		m_last_locker = Thread.getThis.id;
	}
	body {
		m_resource.signal(m_id);
		debug(locker) writef("Locker: %s %s locked\n", Thread.getThis.toString, this.toString);
	}
	
	// manual unlock
	void unlock()
	in {
		assert(m_last_locker == Thread.getThis.id);
	}
	body {
		m_resource.release(m_id);
		debug(locker) writef("Locker: %s %s unlocked\n", Thread.getThis.toString, this.toString);
	}
	
	void set(KT key, VT value, LockerMode mode = LockerMode.AUTO)
	{
		if (mode == LockerMode.AUTO) {
			lock();
		}
		try {
			m_aa.set(key, value);
		} finally {
			if (mode == LockerMode.AUTO) {
				unlock();
			}
		}
	}
	
	VT get(KT key, LockerMode mode = LockerMode.AUTO)
	{
		VT value;
		
		if (mode == LockerMode.AUTO) {
			lock();
		}
		try {
			value = m_aa.get(key);
		} finally {
			if (mode == LockerMode.AUTO) {
				unlock();
			}
		}
		
		return value;
	}
	
	// これは取得中だけしかロックしないので操作中に不整合が起こりうるため
	// 普通は、明示的なlock後にLockerMode.NONで取得して、操作後に明示的にunlockするべし
	VT[] values(LockerMode mode = LockerMode.AUTO)
	{
		VT[] values;
		
		if (mode == LockerMode.AUTO) {
			lock();
		}
		try {
			// aaの実装上valuesはコピーされた配列が戻るため
			// ここではコピーを作らない
			values = m_aa.values;
		} finally {
			if (mode == LockerMode.AUTO) {
				unlock();
			}
		}
		
		return values;
	}
	
	KT[] keys(LockerMode mode = LockerMode.AUTO)
	{
		KT[] keys;
		
		if (mode == LockerMode.AUTO) {
			lock();
		}
		try {
			keys = m_aa.keys;
		} finally {
			if (mode == LockerMode.AUTO) {
				unlock();
			}
		}
		
		return keys;
	}
	
	uint length(LockerMode mode = LockerMode.AUTO)
	{
		uint length;
		
		if (mode == LockerMode.AUTO) {
			lock();
		}
		length = m_aa.length;
		
		if (mode == LockerMode.AUTO) {
			unlock();
		}
		
		return length;
	}
	
	bool exist(KT key, LockerMode mode = LockerMode.AUTO)
	{
		bool found;
		
		if (mode == LockerMode.AUTO) {
			lock();
		}
		try {
			found = m_aa.exist(key);
		} finally {
			if (mode == LockerMode.AUTO) {
				unlock();
			}
		}
		
		return found;
	}
	
	void remove(KT key, LockerMode mode = LockerMode.AUTO)
	{
		if (mode == LockerMode.AUTO) {
			lock();
		}
		try {
			m_aa.remove(key);
		} finally {
			if (mode == LockerMode.AUTO) {
				unlock();
			}
		}
	}
	
	void clear(LockerMode mode = LockerMode.AUTO)
	{
		if (mode == LockerMode.AUTO) {
			lock();
		}
		try {
			foreach (KT e; m_aa.keys) {
				remove(e, LockerMode.NON);
			}
		} finally {
			if (mode == LockerMode.AUTO) {
				unlock();
			}
		}
	}
	
	bool empty(LockerMode mode = LockerMode.AUTO)
	{
		bool is_empty;
		
		if (mode == LockerMode.AUTO) {
			lock();
		}
		
		is_empty = m_aa.length == 0 ? true: false;
		
		if (mode == LockerMode.AUTO) {
			unlock();
		}
		
		
		return is_empty;
	}
	void rehash(LockerMode mode = LockerMode.AUTO)
	{}
/*
	void rehash(LockerMode mode = LockerMode.AUTO)
	{
		
		if (mode == LockerMode.AUTO) {
			lock();
		}
		try {
			m_aa.rehash;
		} finally {
			if (mode == LockerMode.AUTO) {
				unlock();
			}
		}
	}
*/
}

debug(pes_system_locker)
{
	// 正直うまく動いているか微妙
	
	
	import std.stdio;
	import pes.system.eventflag, pes.system.delay;
	
	enum {
		LOCKER_ID 	= 1,
		EVENT_ID	= 2,
		START_EVENT	= 4
	}
	
	
	int main()
	{
		SemaphoreResource resource = new SemaphoreResource;
		Locker!(int, double) locker;
		
		resource.create(LOCKER_ID);
		GlobalEvent.create(EVENT_ID);
		
		locker = new Locker!(int, double)(resource, LOCKER_ID);
		
		Thread t1 = new Thread(&slowtask, locker);
		Thread t2 = new Thread(&task, locker);
		Thread t3 = new Thread(&slowtask, locker);
		Thread t4 = new Thread(&task, locker);
		Thread t5 = new Thread(&slowtask, locker);
		Thread t6 = new Thread(&task, locker);
		Thread t7 = new Thread(&slowtask, locker);
		Thread t8 = new Thread(&task, locker);
		Thread t9 = new Thread(&slowtask, locker);
		
		t1.start();
		t2.start();
		t3.start();
		t4.start();
		t5.start();
		t6.start();
		t7.start();
		t8.start();
		t9.start();
		
		delayTask(1000);
		GlobalEvent.startEvent(EVENT_ID, START_EVENT);
		
		t1.wait();
		t2.wait();
		t3.wait();
		t4.wait();
		t5.wait();
		t6.wait();
		t7.wait();
		t8.wait();
		t9.wait();
		
		return 0;
	}
	
	
	int task(void *dummy)
	{
		Locker!(int, double) locker = cast(Locker!(int, double))dummy;
		uint id = Thread.getThis.id;
		
		GlobalEvent.waitForStartEvent(EVENT_ID, START_EVENT);
		
		for (int i = 1; i <= 4; ++i) {
			locker.set(i, i / 3.14);
		}
		foreach (double e; locker.values.sort) {
			writef("%s-: %s \n", id, e);
		}
		
		foreach (int e; locker.keys.sort) {
			writef("%s-: %s \n", id, locker.get(e));
		}
		
		return 0;
	}
	
	int slowtask(void *dummy)
	{
		Locker!(int, double) locker = cast(Locker!(int, double))dummy;
		uint id = Thread.getThis.id;
		
		GlobalEvent.waitForStartEvent(EVENT_ID, START_EVENT);
		
		for (int i = 1; i <= 4; ++i) {
			locker.set(i, i / 3.14);
		}
		
		foreach (double e; locker.values.sort) {
			writef("%s-: %s \n", id, e);
		}
		
		locker.lock();
		writef("lock!\n");
		delayTask(1000);
		foreach (int e; locker.keys(LockerMode.NON).sort) {
			writef("locking %s-: %s \n", id, locker.get(e, LockerMode.NON));
		}
		writef("unlock!\n");
		locker.unlock();
		
		return 0;
	}
}

