﻿/**
 * Authors: http://d.hatena.ne.jp/ultraist/
 */
module pes.system.eventflag;
/*
 * イベントフラグ
 * 
 * 
 * remark: ビットはAND監視。
 */


// OR監視は、必要になったらモード入れる…。
// mailboxメインで使うとOR監視あまり使わない…


private import 	std.thread;
private import 	pes.system.threadpool,
				pes.system.delay,
				pes.system.task,
				pes.system.synchronizer,
				pes.queue;

public alias	uint	eventflag;

public EventResource	GlobalEvent;

alias	Queue!(Synchronizer)	WaiterQueue;

static this()
{
	GlobalEvent = new EventResource;
}

class EventResource
{
protected:
	EventFlag[eventflag]	m_flag;
	WaiterQueue[eventflag]	m_waiter;
	
public:
	void create(eventflag id)
	{
		synchronized(this) {
			if ((id in m_flag) == null) {
				m_flag[id] 		= new EventFlag;
				m_waiter[id]	= new WaiterQueue;
			}
		}
	}
	
	void remove(eventflag id)
	{
		synchronized(this) {
			m_flag.remove(id);
			while (!m_waiter[id].empty) {
				m_waiter[id].shift.signal;
			}
			m_waiter.remove(id);
		}
	}
	
	void startEvent(eventflag id, uint mask)
	{
		synchronized(this) {
			m_flag[id].setFlag(mask);
			
			while (!m_waiter[id].empty) {
				m_waiter[id].shift.signal;
			}
		}
		
	}
	
	void endEvent(eventflag id, uint mask)
	{
		synchronized(this) {
			m_flag[id].clearFlag(mask);
			
			while (!m_waiter[id].empty) {
				m_waiter[id].shift.signal;
			}
		}
	}
	
	
	void waitForStartEvent(eventflag id, uint event_mask)
	{
		Synchronizer sync = Task.getThis.synchronizer;
		
		while (true) {
			synchronized(this) {
				if ((m_flag[id].getFlag & event_mask) == event_mask) {
					break;
				}
				m_waiter[id].push(sync); //
			}
			sync.wait();
		}
	}
	
	void waitForEndEvent(eventflag id, uint event_mask)
	{
		Synchronizer sync = Task.getThis.synchronizer;
		
		while (true) {
			synchronized(this) {
				if ((m_flag[id].getFlag & event_mask) == 0) {
					break;
				}
				m_waiter[id].push(sync); // ↓
			}
			sync.wait();
		}
	}
	
	// Mutex
	void waitForEndEventAndStart(eventflag id, uint event_mask)
	{
		Synchronizer sync = Task.getThis.synchronizer;
		
		while (true) {
			synchronized(this) {
				if ((m_flag[id].getFlag & event_mask) == 0) {
					m_flag[id].setFlag(event_mask);
					break;
				}
				m_waiter[id].push(sync); // ↓
			}
			sync.wait();
		}
		//delayTask(10);
	}
	
	uint referEvent(eventflag id)
	{
		synchronized(this) {
			return m_flag[id].getFlag;
		}
	}
	
	// 10msec単位で監視する
	bool pollStartEvent(eventflag id, uint event_mask, uint wait_msec = 0)
	{
		bool on = false;
		
		while (true) {
			synchronized(this) {
				if ((m_flag[id].getFlag & event_mask) == event_mask) {
					on = true;
					break;
				}
			}
			if (wait_msec == 0) {
				break;
			} else if (wait_msec > 10) {
				delayTask(10);
				wait_msec -= 10;
			} else {
				delayTask(wait_msec);
				wait_msec  = 0;
			}
		}
		
		return on;
	}
	
	// 10msec単位で監視する
	bool pollEndEvent(eventflag id, uint event_mask, uint wait_msec = 0)
	{
		bool on = false;
		
		while (true) {
			synchronized(this) {
				if ((m_flag[id].getFlag & event_mask) == 0) {
					on = true;
					break;
				}
			}
			if (wait_msec == 0) {
				break;
			} else if (wait_msec > 10) {
				delayTask(10);
				wait_msec -= 10;
			} else {
				delayTask(wait_msec);
				wait_msec  = 0;
			}
		}
		
		return on;
	}
	
	
	// 10msec単位で監視する
	bool pollEndEventAndStart(eventflag id, uint event_mask, uint wait_msec = 0)
	{
		bool on = false;
		
		while (true) {
			synchronized(this) {
				if ((m_flag[id].getFlag & event_mask) == 0) {
					m_flag[id].setFlag(event_mask);
					on = true;
					break;
				}
			}
			if (wait_msec == 0) {
				break;
			} else if (wait_msec > 10) {
				delayTask(10);
				wait_msec -= 10;
			} else {
				delayTask(wait_msec);
				wait_msec  = 0;
			}
		}
		
		return on;
	}
	
	eventflag[] ids()
	{
		return m_flag.keys;
	}
}


class EventFlag
{
protected:
	uint	m_flag;
public:
	this()
	{
		m_flag = 0;
	}
	
	uint getFlag()
	{
		return m_flag;
	}
	
	void setFlag(uint mask)
	{
		m_flag |= mask;
	}
	
	void clearFlag(uint mask)
	{
		m_flag &= ~mask;
	}
}


debug(pes_system_eventflag)
{
	import std.stdio;
	import std.thread;
	import std.c.time;
	
	enum: eventflag {
		EVENT_ID = 1
	}
	
	enum: uint {
		EVENT1 = 1,
		EVENT2 = 2,
		EVENT3 = 4
	}
	
	int main()
	{
		EventResource resource = new EventResource;
		resource.create(EVENT1);
		resource.create(EVENT2);
		resource.create(EVENT3);
		
		Thread t1 = new Thread(&task1, resource);
		Thread t2 = new Thread(&task2, resource);
		
		t1.start();
		t2.start();
		
		writef("main: start event1\n");
		resource.startEvent(EVENT_ID, EVENT1);
		sleep(2);
		writef("main: start event3\n");
		resource.startEvent(EVENT_ID, EVENT3);
		
		resource.waitForEndEvent(EVENT_ID, EVENT1);
		writef("main: resume end event1\n");
		writef("main: end event2\n");
		resource.endEvent(EVENT_ID, EVENT2);
		writef("main: end event3\n");
		resource.endEvent(EVENT_ID, EVENT3);
		
		t1.wait();
		t2.wait();
		
		return 0;
	}
	
	int task1(void *dummy)
	{
		EventResource resource = cast(EventResource)dummy;
		
		resource.waitForStartEvent(EVENT_ID, EVENT1);
		writef("task1: resume start event1\n");
		resource.waitForEndEvent(EVENT_ID, EVENT1 | EVENT2 | EVENT3);
		writef("task1: resume start event1,2,3\n");
		
		return 0;
	}
	
	int task2(void *dummy)
	{
		EventResource resource = cast(EventResource)dummy;
		
		writef("task2: start event2\n");
		resource.startEvent(EVENT_ID, EVENT2);
		
		resource.waitForStartEvent(EVENT_ID, EVENT2);
		writef("task2: resume start event2\n");
		
		
		resource.waitForStartEvent(EVENT_ID, EVENT3);
		writef("task2: resume start event3\n");
		
		writef("task2: end event1\n");
		resource.endEvent(EVENT_ID, EVENT1);
		
		return 0;
	}
}
